@@ -62,12 +62,9 @@ const Settings = imports.ui.settings;
 const SignalManager = imports.misc.signalManager;
 const Tooltips = imports.ui.tooltips;
 
-const HORIZONTAL_ICON_SIZE = 16; // too bad this can't be defined in theme (cinnamon-app.create_icon_texture returns a clutter actor, not a themable object -
-                                 // probably something that could be addressed
-const ICON_HEIGHT_FACTOR = 0.64;
-const VERTICAL_ICON_HEIGHT_FACTOR = 0.75;
 const MAX_TEXT_LENGTH = 1000;
 const FLASH_INTERVAL = 500;
+const FLASH_MAX_COUNT = 4;
 
 const WINDOW_PREVIEW_WIDTH = 200;
 const WINDOW_PREVIEW_HEIGHT = 150;
@@ -76,11 +73,14 @@ class WindowPreview extends Tooltips.TooltipBase {
     constructor(item, metaWindow, previewScale, showLabel) {
         super(item.actor);
         this._applet = item._applet;
-        this.uiScale = St.ThemeContext.get_for_stage(global.stage).scale_factor;
-        this.thumbScale = previewScale;
         this.metaWindow = metaWindow;
-        this.muffinWindow = null;
-        this._sizeChangedId = null;
+        this._windowActor = null;
+        let x = this.windowActor;
+        this.uiScale = global.ui_scale;
+        this.thumbScale = previewScale;
+
+        this._sizeChangedId = 0;
+        this.thumbnail = null;
 
         this.actor = new St.BoxLayout({ vertical: true, style_class: "window-list-preview", important: true });
         this.actor.show_on_set_parent = false;
@@ -100,6 +100,21 @@ class WindowPreview extends Tooltips.TooltipBase {
         this.actor.add_actor(this.thumbnailBin);
     }
 
+    get windowActor() {
+        if (this._windowActor) {
+            return this._windowActor;
+        }
+
+        this._windowActor = this.metaWindow.get_compositor_private();
+
+        if (this._windowActor) {
+            return this._windowActor;
+        } else {
+            log("metaWindow has no actor!");
+            return null;
+        }
+    }
+
     _onEnterEvent(actor, event) {
         if (this._applet._tooltipShowing)
             this.show();
@@ -118,7 +133,7 @@ class WindowPreview extends Tooltips.TooltipBase {
     }
 
     _hide(actor, event) {
-        Tooltips.TooltipBase.prototype._hide.call(this, actor, event);
+        super._hide.call(this, actor, event);
         this._applet.erodeTooltip();
     }
 
@@ -126,53 +141,86 @@ class WindowPreview extends Tooltips.TooltipBase {
         if (!this.actor || this._applet._menuOpen)
             return;
 
-        this.muffinWindow = this.metaWindow.get_compositor_private();
-        let windowTexture = this.muffinWindow.get_texture();
-        let [width, height] = this._getScaledTextureSize(windowTexture);
-
         if (this.thumbnail) {
             this.thumbnailBin.set_child(null);
             this.thumbnail.destroy();
+            this.thumbnail = null;
         }
 
+        let windowTexture = this.windowActor.get_texture();
+
+        if (!windowTexture) {
+            this.actor.hide();
+            return;
+        }
+
+        let [width, height] = this._getScaledTextureSize(windowTexture);
+
         this.thumbnail = new Clutter.Clone({
             source: windowTexture,
             width: width,
             height: height
         });
 
-        this._sizeChangedId = this.muffinWindow.connect('size-changed',
-            Lang.bind(this, function() {
-                let [width, height] = this._getScaledTextureSize(windowTexture);
-                this.thumbnail.set_size(width, height);
-            }));
+        this._sizeChangedId = this.windowActor.connect('size-changed', () => {
+            let [width, height] = this._getScaledTextureSize(windowTexture);
+            this.thumbnail.set_size(width, height);
+            this._set_position();
+        });
 
         this.thumbnailBin.set_child(this.thumbnail);
 
+        this.actor.show();
+        this._set_position();
+
+        this.visible = true;
+        this._applet.cancelErodeTooltip();
+        this._applet._tooltipShowing = true;
+    }
+
+    hide() {
+        if (this._sizeChangedId > 0) {
+            this.windowActor.disconnect(this._sizeChangedId);
+            this._sizeChangedId = 0;
+        }
+        if (this.thumbnail) {
+            this.thumbnailBin.set_child(null);
+            this.thumbnail.destroy();
+            this.thumbnail = null;
+        }
+        if (this.actor) {
+            this.actor.hide();
+        }
+        this.visible = false;
+    }
+
+    _set_position() {
+        if (!this.actor || this.actor.is_finalized()) return;
         let allocation = this.actor.get_allocation_box();
         let previewHeight = allocation.y2 - allocation.y1;
         let previewWidth = allocation.x2 - allocation.x1;
 
         let monitor = Main.layoutManager.findMonitorForActor(this.item);
-        let previewTop;
 
-        if (this._applet.orientation == St.Side.BOTTOM) {
+        let previewTop;
+        if (this._applet.orientation === St.Side.BOTTOM) {
             previewTop = this.item.get_transformed_position()[1] - previewHeight - 5;
-        } else if (this._applet.orientation == St.Side.TOP) {
+        } else if (this._applet.orientation === St.Side.TOP) {
             previewTop = this.item.get_transformed_position()[1] + this.item.get_transformed_size()[1] + 5;
         } else {
             previewTop = this.item.get_transformed_position()[1];
         }
 
         let previewLeft;
-        if (this._applet.orientation == St.Side.BOTTOM || this._applet.orientation == St.Side.TOP) {
+        if (this._applet.orientation === St.Side.BOTTOM || this._applet.orientation === St.Side.TOP) {
             // centre the applet on the window list item if window list is on the top or bottom panel
             previewLeft = this.item.get_transformed_position()[0] + this.item.get_transformed_size()[0]/2 - previewWidth/2;
-        } else if (this._applet.orientation == St.Side.LEFT) {
+        } else if (this._applet.orientation === St.Side.LEFT) {
             previewLeft = this.item.get_transformed_position()[0] + this.item.get_transformed_size()[0] + 5;
-        } else if (this._applet.orientation == St.Side.RIGHT) {
-            previewLeft = this.item.get_transformed_position()[0] - previewWidth -5;
+        } else {
+            previewLeft = this.item.get_transformed_position()[0] - previewWidth - 5;
         }
+
         previewLeft = Math.round(previewLeft);
         previewLeft = Math.max(previewLeft, monitor.x);
         previewLeft = Math.min(previewLeft, monitor.x + monitor.width - previewWidth);
@@ -181,26 +229,6 @@ class WindowPreview extends Tooltips.TooltipBase {
         previewTop  = Math.min(previewTop, monitor.y + monitor.height - previewHeight);
 
         this.actor.set_position(previewLeft, previewTop);
-
-        this.actor.show();
-        this.visible = true;
-        this._applet.cancelErodeTooltip();
-        this._applet._tooltipShowing = true;
-    }
-
-    hide() {
-        if (this._sizeChangedId != null) {
-            this.muffinWindow.disconnect(this._sizeChangedId);
-            this._sizeChangedId = null;
-        }
-        if (this.thumbnail) {
-            this.thumbnailBin.set_child(null);
-            this.thumbnail.destroy();
-        }
-        if (this.actor) {
-            this.actor.hide();
-        }
-        this.visible = false;
     }
 
     set_text(text) {
@@ -208,18 +236,20 @@ class WindowPreview extends Tooltips.TooltipBase {
     }
 
     _destroy() {
-        if (this._sizeChangedId != null) {
-            this.muffinWindow.disconnect(this._sizeChangedId);
-            this.sizeChangedId = null;
+        if (this._sizeChangedId > 0) {
+            this._windowActor.disconnect(this._sizeChangedId);
+            this.sizeChangedId = 0;
         }
         if (this.thumbnail) {
             this.thumbnailBin.set_child(null);
             this.thumbnail.destroy();
+            this.thumbnail = null;
         }
         if (this.actor) {
+            Main.uiGroup.remove_actor(this.actor);
             this.actor.destroy();
+            this.actor = null;
         }
-        this.actor = null;
     }
 }
 
@@ -234,9 +264,10 @@ class AppMenuButton {
 
         this._applet = applet;
         this.metaWindow = metaWindow;
-        this.alert = alert;
+        this.alert = alert || metaWindow.is_demanding_attention() || metaWindow.is_urgent();
         this.labelVisible = false;
-        this.window_signals = new SignalManager.SignalManager(null);
+        this._signals = new SignalManager.SignalManager();
+        this.xid = metaWindow.get_xwindow();
 
         if (this._applet.orientation == St.Side.TOP)
             this.actor.add_style_class_name('top');
@@ -248,23 +279,23 @@ class AppMenuButton {
             this.actor.add_style_class_name('right');
 
         this.actor._delegate = this;
-        this.actor.connect('button-release-event', Lang.bind(this, this._onButtonRelease));
-        this.actor.connect('button-press-event', Lang.bind(this, this._onButtonPress));
+        this._signals.connect(this.actor, 'button-release-event', Lang.bind(this, this._onButtonRelease));
+        this._signals.connect(this.actor, 'button-press-event', Lang.bind(this, this._onButtonPress));
 
-        this.actor.connect('get-preferred-width',
+        this._signals.connect(this.actor, 'get-preferred-width',
                 Lang.bind(this, this._getPreferredWidth));
-        this.actor.connect('get-preferred-height',
+        this._signals.connect(this.actor, 'get-preferred-height',
                 Lang.bind(this, this._getPreferredHeight));
-        this.actor.connect('allocate', Lang.bind(this, this._allocate));
+        this._signals.connect(this.actor, 'allocate', Lang.bind(this, this._allocate));
 
         this.progressOverlay = new St.Widget({ style_class: "progress", reactive: false, important: true  });
 
         this.actor.add_actor(this.progressOverlay);
 
         this._iconBox = new Cinnamon.Slicer({ name: 'appMenuIcon' });
-        this._iconBox.connect('style-changed',
+        this._signals.connect(this._iconBox, 'style-changed',
                               Lang.bind(this, this._onIconBoxStyleChanged));
-        this._iconBox.connect('notify::allocation',
+        this._signals.connect(this._iconBox, 'notify::allocation',
                               Lang.bind(this, this._updateIconBoxClipAndGeometry));
         this.actor.add_actor(this._iconBox);
 
@@ -311,13 +342,13 @@ class AppMenuButton {
             this._menuManager.addMenu(this.rightClickMenu);
 
             this._draggable = DND.makeDraggable(this.actor, null, this._applet.actor);
-            this._draggable.connect('drag-begin', Lang.bind(this, this._onDragBegin));
-            this._draggable.connect('drag-cancelled', Lang.bind(this, this._onDragCancelled));
-            this._draggable.connect('drag-end', Lang.bind(this, this._onDragEnd));
+            this._signals.connect(this._draggable, 'drag-begin', Lang.bind(this, this._onDragBegin));
+            this._signals.connect(this._draggable, 'drag-cancelled', Lang.bind(this, this._onDragCancelled));
+            this._signals.connect(this._draggable, 'drag-end', Lang.bind(this, this._onDragEnd));
         }
 
         this.onPanelEditModeChanged();
-        global.settings.connect('changed::panel-edit-mode', Lang.bind(this, this.onPanelEditModeChanged));
+        this._signals.connect(global.settings, 'changed::panel-edit-mode', Lang.bind(this, this.onPanelEditModeChanged));
 
         this._windows = this._applet._windows;
 
@@ -332,12 +363,12 @@ class AppMenuButton {
         if (this.alert)
             this.getAttention();
 
-        this.window_signals.connect(this.metaWindow, 'notify::title', this.setDisplayTitle, this);
-        this.window_signals.connect(this.metaWindow, "notify::minimized", this.setDisplayTitle, this);
-        this.window_signals.connect(this.metaWindow, "notify::tile-type", this.setDisplayTitle, this);
-        this.window_signals.connect(this.metaWindow, "notify::icon", this.setIcon, this);
-        this.window_signals.connect(this.metaWindow, "notify::appears-focused", this.onFocus, this);
-        this.window_signals.connect(this.metaWindow, "unmanaged", this.onUnmanaged, this);
+        this._signals.connect(this.metaWindow, 'notify::title', this.setDisplayTitle, this);
+        this._signals.connect(this.metaWindow, "notify::minimized", this.setDisplayTitle, this);
+        this._signals.connect(this.metaWindow, "notify::tile-type", this.setDisplayTitle, this);
+        this._signals.connect(this.metaWindow, "icon-changed", this.setIcon, this);
+        this._signals.connect(this.metaWindow, "notify::appears-focused", this.onFocus, this);
+        this._signals.connect(this.metaWindow, "unmanaged", this.onUnmanaged, this);
     }
 
     onUnmanaged() {
@@ -496,10 +527,11 @@ class AppMenuButton {
     }
 
     destroy() {
-        this.window_signals.disconnectAllSignals();
+        this._signals.disconnectAllSignals();
         this._tooltip.destroy();
-        if (this.rightClickMenu) {
+        if (!this.alert) {
             this.rightClickMenu.destroy();
+            this._menuManager.destroy();
         }
         this.actor.destroy();
     }
@@ -573,7 +605,7 @@ class AppMenuButton {
         if (!this._hasFocus()) {
             Main.activateWindow(this.metaWindow, global.get_current_time());
             this.actor.add_style_pseudo_class('focus');
-        } else if (!fromDrag) {
+        } else if (!fromDrag && this._applet.leftClickMinimize) {
             this.metaWindow.minimize();
             this.actor.remove_style_pseudo_class('focus');
         }
@@ -667,7 +699,7 @@ class AppMenuButton {
             }
             childBox.x2 = Math.min(childBox.x1 + naturalWidth, box.x2);
         } else {
-            childBox.x1 = box.x1 + Math.max(0, (allocWidth - naturalWidth) / 2);
+            childBox.x1 = box.x1 + Math.floor(Math.max(0, allocWidth - naturalWidth) / 2);
             childBox.x2 = Math.min(childBox.x1 + naturalWidth, box.x2);
         }
         this._iconBox.allocate(childBox, flags);
@@ -719,18 +751,13 @@ class AppMenuButton {
         let tracker = Cinnamon.WindowTracker.get_default();
         let app = tracker.get_window_app(this.metaWindow);
 
-        if (this._applet._scaleMode && this.labelVisible)
-            this.iconSize = Math.round(this._applet._panelHeight * ICON_HEIGHT_FACTOR / global.ui_scale);
-        else if (!this.labelVisible)
-            this.iconSize = Math.round(this._applet._panelHeight * VERTICAL_ICON_HEIGHT_FACTOR / global.ui_scale);
-        else
-            this.iconSize = HORIZONTAL_ICON_SIZE;
+        this.icon_size = this._applet.icon_size;
 
         let icon = app ?
-            app.create_icon_texture_for_window(this.iconSize, this.metaWindow) :
+            app.create_icon_texture_for_window(this.icon_size, this.metaWindow) :
             new St.Icon({ icon_name: 'application-default-icon',
                 icon_type: St.IconType.FULLCOLOR,
-                icon_size: this.iconSize });
+                icon_size: this.icon_size });
 
         let old_child = this._iconBox.get_child();
         this._iconBox.set_child(icon);
@@ -744,26 +771,28 @@ class AppMenuButton {
             return false;
 
         this._needsAttention = true;
-        let counter = 0;
-        this._flashButton(counter);
+        this._flashButton();
         return true;
     }
 
-    _flashButton(counter) {
+    _flashButton() {
         if (!this._needsAttention)
             return;
 
-        this.actor.add_style_class_name("window-list-item-demands-attention");
-        if (counter < 4) {
-            Mainloop.timeout_add(FLASH_INTERVAL, Lang.bind(this, function () {
-                if (this.actor.has_style_class_name("window-list-item-demands-attention")) {
-                    this.actor.remove_style_class_name("window-list-item-demands-attention");
-                }
-                Mainloop.timeout_add(FLASH_INTERVAL, Lang.bind(this, function () {
-                    this._flashButton(++counter);
-                }));
-            }));
-        }
+        let counter = 0;
+        let sc = "window-list-item-demands-attention";
+
+        Mainloop.timeout_add(FLASH_INTERVAL, () => {
+            if (!this._needsAttention)
+                return false;
+
+            if (this.actor.has_style_class_name(sc))
+                this.actor.remove_style_class_name(sc);
+            else
+                this.actor.add_style_class_name(sc);
+
+            return counter++ < FLASH_MAX_COUNT;
+        });
     }
 };
 
@@ -773,7 +802,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
 
         this._launcher = launcher;
         this._windows = launcher._applet._windows;
-        this.connect('open-state-changed', Lang.bind(this, this._onToggled));
+        this._signals.connect(this, 'open-state-changed', Lang.bind(this, this._onToggled));
 
         this.orientation = orientation;
         this.metaWindow = metaWindow;
@@ -789,7 +818,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
             Main.layoutManager.monitors.forEach(function (monitor, index) {
                 if (index === mw.get_monitor()) return;
                 item = new PopupMenu.PopupMenuItem(_("Move to the other monitor"));
-                item.connect('activate', function() {
+                this._signals.connect(item, 'activate', function() {
                     mw.move_to_monitor(index);
                 });
                 this.addMenuItem(item);
@@ -799,7 +828,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
             Main.layoutManager.monitors.forEach(function (monitor, index) {
                 if (index === mw.get_monitor()) return;
                 item = new PopupMenu.PopupMenuItem(_("Move to monitor %d").format(index + 1));
-                item.connect('activate', function() {
+                this._signals.connect(item, 'activate', function() {
                     mw.move_to_monitor(index);
                 });
                 this.addMenuItem(item);
@@ -810,13 +839,13 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
         if ((length = global.screen.n_workspaces) > 1) {
             if (mw.is_on_all_workspaces()) {
                 item = new PopupMenu.PopupMenuItem(_("Only on this workspace"));
-                item.connect('activate', function() {
+                this._signals.connect(item, 'activate', function() {
                     mw.unstick();
                 });
                 this.addMenuItem(item);
             } else {
                 item = new PopupMenu.PopupMenuItem(_("Visible on all workspaces"));
-                item.connect('activate', function() {
+                this._signals.connect(item, 'activate', function() {
                     mw.stick();
                 });
                 this.addMenuItem(item);
@@ -834,7 +863,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
                     if (i == curr_index)
                         ws.setSensitive(false);
 
-                    ws.connect('activate', function() {
+                    this._signals.connect(ws, 'activate', function() {
                         mw.change_workspace(global.screen.get_workspace_by_index(j));
                     });
                     item.menu.addMenuItem(ws);
@@ -848,24 +877,22 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
         this.addMenuItem(subMenu);
 
         item = new PopupMenu.PopupIconMenuItem(_("About..."), "dialog-question", St.IconType.SYMBOLIC);
-        item.connect('activate', Lang.bind(this._launcher._applet, this._launcher._applet.openAbout));
+        this._signals.connect(item, 'activate', Lang.bind(this._launcher._applet, this._launcher._applet.openAbout));
         subMenu.menu.addMenuItem(item);
 
         item = new PopupMenu.PopupIconMenuItem(_("Configure..."), "system-run", St.IconType.SYMBOLIC);
-        item.connect('activate', Lang.bind(this._launcher._applet, this._launcher._applet.configureApplet));
+        this._signals.connect(item, 'activate', Lang.bind(this._launcher._applet, this._launcher._applet.configureApplet));
         subMenu.menu.addMenuItem(item);
 
         item = new PopupMenu.PopupIconMenuItem(_("Remove 'Window list'"), "edit-delete", St.IconType.SYMBOLIC);
-        item.connect('activate', Lang.bind(this, function() {
-            AppletManager._removeAppletFromPanel(this._launcher._applet._uuid, this._launcher._applet.instance_id);
-        }));
+        this._signals.connect(item, 'activate', (actor, event) => this._launcher._applet.confirmRemoveApplet(event));
         subMenu.menu.addMenuItem(item);
 
         this.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
 
         // Close all/others
         item = new PopupMenu.PopupIconMenuItem(_("Close all"), "application-exit", St.IconType.SYMBOLIC);
-        item.connect('activate', Lang.bind(this, function() {
+        this._signals.connect(item, 'activate', Lang.bind(this, function() {
             for (let window of this._windows)
                 if (window.actor.visible &&
                    !window._needsAttention)
@@ -874,7 +901,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
         this.addMenuItem(item);
 
         item = new PopupMenu.PopupIconMenuItem(_("Close others"), "window-close", St.IconType.SYMBOLIC);
-        item.connect('activate', Lang.bind(this, function() {
+        this._signals.connect(item, 'activate', Lang.bind(this, function() {
             for (let window of this._windows)
                 if (window.actor.visible &&
                     window.metaWindow != this.metaWindow &&
@@ -888,7 +915,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
         // Miscellaneous
         if (mw.get_compositor_private().opacity != 255) {
             item = new PopupMenu.PopupMenuItem(_("Restore to full opacity"));
-            item.connect('activate', function() {
+            this._signals.connect(item, 'activate', function() {
                 mw.get_compositor_private().set_opacity(255);
             });
             this.addMenuItem(item);
@@ -896,12 +923,12 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
 
         if (mw.minimized) {
             item = new PopupMenu.PopupIconMenuItem(_("Restore"), "view-sort-descending", St.IconType.SYMBOLIC);
-            item.connect('activate', function() {
+            this._signals.connect(item, 'activate', function() {
                 Main.activateWindow(mw, global.get_current_time());
             });
         } else {
             item = new PopupMenu.PopupIconMenuItem(_("Minimize"), "view-sort-ascending", St.IconType.SYMBOLIC);
-            item.connect('activate', function() {
+            this._signals.connect(item, 'activate', function() {
                 mw.minimize();
             });
         }
@@ -909,19 +936,20 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
 
         if (mw.get_maximized()) {
             item = new PopupMenu.PopupIconMenuItem(_("Unmaximize"), "view-restore", St.IconType.SYMBOLIC);
-            item.connect('activate', function() {
+            this._signals.connect(item, 'activate', function() {
                 mw.unmaximize(Meta.MaximizeFlags.HORIZONTAL | Meta.MaximizeFlags.VERTICAL);
             });
         } else {
             item = new PopupMenu.PopupIconMenuItem(_("Maximize"), "view-fullscreen", St.IconType.SYMBOLIC);
-            item.connect('activate', function() {
+            this._signals.connect(item, 'activate', function() {
                 mw.maximize(Meta.MaximizeFlags.HORIZONTAL | Meta.MaximizeFlags.VERTICAL);
             });
         }
         this.addMenuItem(item);
+        item.setSensitive(mw.can_maximize())
 
         item = new PopupMenu.PopupIconMenuItem(_("Close"), "edit-delete", St.IconType.SYMBOLIC);
-        item.connect('activate', function() {
+        this._signals.connect(item, 'activate', function() {
             mw.delete(global.get_current_time());
         });
         this.addMenuItem(item);
@@ -955,6 +983,7 @@ class CinnamonWindowListApplet extends Applet.Applet {
         this.actor.set_track_hover(false);
         this.actor.set_style_class_name("window-list-box");
         this.orientation = orientation;
+        this.icon_size = this.getPanelIconSize(St.IconType.FULLCOLOR);
         this.appletEnabled = false;
         //
         // A layout manager is used to cater for vertical panels as well as horizontal
@@ -985,11 +1014,13 @@ class CinnamonWindowListApplet extends Applet.Applet {
         this.settings.bind("enable-alerts", "enableAlerts", this._updateAttentionGrabber);
         this.settings.bind("enable-scrolling", "scrollable", this._onEnableScrollChanged);
         this.settings.bind("reverse-scrolling", "reverseScroll");
+        this.settings.bind("left-click-minimize", "leftClickMinimize");
         this.settings.bind("middle-click-close", "middleClickClose");
         this.settings.bind("buttons-use-entire-space", "buttonsUseEntireSpace", this._refreshAllItems);
         this.settings.bind("window-preview", "usePreview", this._onPreviewChanged);
         this.settings.bind("window-preview-show-label", "showLabel", this._onPreviewChanged);
         this.settings.bind("window-preview-scale", "previewScale", this._onPreviewChanged);
+        this.settings.bind("last-window-order", "lastWindowOrder", null);
 
         this.signals.connect(global.screen, 'window-added', this._onWindowAddedAsync, this);
         this.signals.connect(global.screen, 'window-monitor-changed', this._onWindowMonitorChanged, this);
@@ -997,8 +1028,9 @@ class CinnamonWindowListApplet extends Applet.Applet {
         this.signals.connect(global.screen, 'window-skip-taskbar-changed', this._onWindowSkipTaskbarChanged, this);
         this.signals.connect(global.screen, 'monitors-changed', this._updateWatchedMonitors, this);
         this.signals.connect(global.window_manager, 'switch-workspace', this._refreshAllItems, this);
+        this.signals.connect(Cinnamon.WindowTracker.get_default(), "window-app-changed", this._onWindowAppChanged, this);
 
-        this.actor.connect('style-changed', Lang.bind(this, this._updateSpacing));
+        this.signals.connect(this.actor, 'style-changed', Lang.bind(this, this._updateSpacing));
 
         global.settings.bind("panel-edit-mode", this.actor, "reactive", Gio.SettingsBindFlags.DEFAULT);
 
@@ -1013,6 +1045,7 @@ class CinnamonWindowListApplet extends Applet.Applet {
 
     on_applet_removed_from_panel() {
         this.signals.disconnectAllSignals();
+        this.settings.finalize();
     }
 
     on_applet_instances_changed() {
@@ -1020,6 +1053,12 @@ class CinnamonWindowListApplet extends Applet.Applet {
     }
 
     on_panel_height_changed() {
+        this.icon_size = this.getPanelIconSize(St.IconType.FULLCOLOR);
+        this._refreshAllItems();
+    }
+
+    on_panel_icon_size_changed(size) {
+        this.icon_size = size;
         this._refreshAllItems();
     }
 
@@ -1098,17 +1137,23 @@ class CinnamonWindowListApplet extends Applet.Applet {
             this._removeWindow(metaWindow);
     }
 
-    _onWindowWorkspaceChanged(screen, metaWindow, metaWorkspace) {
+    _refreshItemByMetaWindow(metaWindow) {
         let window = this._windows.find(win => (win.metaWindow == metaWindow));
 
         if (window)
             this._refreshItem(window);
     }
 
-    _onWindowSkipTaskbarChanged(screen, metaWindow) {
-        let window = this._windows.find(win => (win.metaWindow == metaWindow));
+    _onWindowWorkspaceChanged(screen, metaWindow, metaWorkspace) {
+        this._refreshItemByMetaWindow(metaWindow);
+    }
 
-        if (window && !Main.isInteresting(metaWindow)) {
+    _onWindowAppChanged(tracker, metaWindow) {
+        this._refreshItemByMetaWindow(metaWindow);
+    }
+
+    _onWindowSkipTaskbarChanged(screen, metaWindow) {
+        if (metaWindow && metaWindow.is_skip_taskbar()) {
             this._removeWindow(metaWindow);
             return;
         }
@@ -1165,6 +1210,9 @@ class CinnamonWindowListApplet extends Applet.Applet {
          * one isn't shown! */
         if (window.alert)
             window.actor.visible = !window.actor.visible;
+
+        if (window.actor.visible)
+            window.setIcon();
     }
 
     _refreshAllItems() {
@@ -1218,6 +1266,7 @@ class CinnamonWindowListApplet extends Applet.Applet {
             }
         }
 
+        this.refreshing = true;
 
         for (let window of windows) {
             if (this._shouldAdd(window))
@@ -1225,6 +1274,10 @@ class CinnamonWindowListApplet extends Applet.Applet {
             else
                 this._removeWindow(window);
         }
+
+        this.refreshing = false;
+
+        this._applySavedOrder();
     }
 
     _addWindow(metaWindow, alert) {
@@ -1252,6 +1305,8 @@ class CinnamonWindowListApplet extends Applet.Applet {
                 }
             }
         }
+
+        this._saveOrder();
     }
 
     _removeWindow(metaWindow) {
@@ -1263,13 +1318,62 @@ class CinnamonWindowListApplet extends Applet.Applet {
                 this._windows.splice(i, 1);
             }
         }
+
+        this._saveOrder();
     }
 
     _shouldAdd(metaWindow) {
         return Main.isInteresting(metaWindow) &&
+            !metaWindow.is_skip_taskbar() &&
             this._monitorWatchList.indexOf(metaWindow.get_monitor()) != -1;
     }
 
+    /* Store by Windows (XIDs), a simple list
+       xid::xid::xid::xid::xid
+    */
+
+    _applySavedOrder() {
+        let order = this.lastWindowOrder.split("::");
+
+        order.reverse();
+
+        for (let i = 0; i < order.length; i++) {
+            let xid = parseInt(order[i]);
+
+            if (xid === NaN) {
+                continue;
+            }
+
+            let found = this._windows.find(win => (win.xid == xid));
+
+            if (found) {
+                this.manager_container.set_child_at_index(found.actor, 0);
+            }
+        }
+
+        this._saveOrder();
+    }
+
+    _saveOrder() {
+        if (this.refreshing) {
+            return;
+        }
+
+        let new_order = [];
+        let actors = this.manager_container.get_children();
+
+        for (let i = 0; i < actors.length; i++) {
+            new_order.push(actors[i]._delegate.xid);
+        }
+
+        if (new_order.length === 0) {
+            this.lastWindowOrder = "";
+            return;
+        }
+
+        this.lastWindowOrder = new_order.join("::");
+    }
+
     handleDragOver(source, actor, x, y, time) {
         if (this._inEditMode)
             return DND.DragMotionResult.MOVE_DROP;
@@ -1278,20 +1382,17 @@ class CinnamonWindowListApplet extends Applet.Applet {
 
         let children = this.manager_container.get_children();
         let isVertical = this.manager_container.height > this.manager_container.width;
+        let axis = isVertical ? [y, 'y1'] : [x, 'x1'];
 
-        this._dragPlaceholderPos = -1
-        for (let i = children.length - 1; i >= 0; i--) {
+        this._dragPlaceholderPos = -1;
+        let minDist = -1;
+        for(let i = 0; i < children.length; i++) {
             if (!children[i].visible)
                 continue;
-
-            if (isVertical) {
-                if (y > children[i].get_allocation_box().y1) {
-                    this._dragPlaceholderPos = i;
-                    break;
-                }
-            } else if  (x > children[i].get_allocation_box().x1) {
+            let dist = Math.abs(axis[0] - (children[i].get_allocation_box()[axis[1]] + children[i].width / 2));
+            if(dist < minDist || minDist == -1) {
+                minDist = dist;
                 this._dragPlaceholderPos = i;
-                break;
             }
         }
 
@@ -1317,6 +1418,8 @@ class CinnamonWindowListApplet extends Applet.Applet {
 
         this.manager_container.set_child_at_index(source.actor, this._dragPlaceholderPos);
 
+        this._saveOrder();
+
         return true;
     }
 
